﻿using NetTopologySuite.Geometries;
using SkiaSharp;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Input;
using ThinkGeo.Core;
using ThinkGeo.UI.XamarinForms;
using Xamarin.Essentials;
using Xamarin.Forms;

namespace GoogleCache
{
    public class MainPageViewModel : INotifyPropertyChanged
    {
        MapView mapView;
        public ICommand DownloadCommand { get; set; }
        private string buttonText;

        public string ButtonText
        {
            get { return buttonText; }
            set { buttonText = value; OnPropertyChanged(nameof(ButtonText)); }
        }
        private bool isEnabled;

        public bool IsEnabled
        {
            get { return isEnabled; }
            set { isEnabled = value; OnPropertyChanged(nameof(IsEnabled)); }
        }


        public MainPageViewModel(MapView map)
        {
            DownloadCommand = new Command(ExecuteTileCaching);
            mapView = map;
            SetupMap();
            ButtonText = "Download Current Extent Cache";
            IsEnabled = true;
        }


        public void SetupMap() 
        {
            MapHelper.InitilizeMapView(mapView);
            MapHelper.SetupGoogleMapOverlay(mapView);
        }

        private async void ExecuteTileCaching()
        {
            ButtonText = "Downloading Data";
            IsEnabled = false;
            RectangleShape currentExtent = mapView.CurrentExtent;
            double zoomLevel = mapView.MapScale;
     
            await Task.Factory.StartNew(() =>
            {
                string cacheDirectory = MapHelper.GetTileCacheFolder();
                string cacheId = "google";

                var currentScale = mapView.MapScale;
                List<double> minAndMaxScales = new List<double>();
                List<(double, int)> minAndMaxScalesWithZoom = new List<(double, int)>();
                var zoomLevels = mapView.ZoomLevelSet.GetZoomLevels();

                var currentZoomLevel = zoomLevels.Where(x => x.Scale >= currentScale).OrderBy(y => y.Scale).FirstOrDefault();
                var zoomLevelIndex = zoomLevels.IndexOf(currentZoomLevel);
                for (int i = zoomLevelIndex-2; i < (zoomLevelIndex + 2); i++)
                {
                    minAndMaxScales.Add(zoomLevels[i].Scale);
                    minAndMaxScalesWithZoom.Add((zoomLevels[i].Scale, i));
                }

                Collection<Layer> layersToCache = new Collection<Layer>();
                GoogleMapsLayer layer = new GoogleMapsLayer();
                layer.IsVisible = true;
                layer.SendingWebRequest += Layer_SendingWebRequest;
                layer.MapType = GoogleMapsMapType.Hybrid;
                layersToCache.Add(layer);
                var extents = mapView.CurrentExtent.ToString().Split(',');
                RectangleShape restrictedExtent = new RectangleShape(double.Parse(extents[0]), double.Parse(extents[1]), double.Parse(extents[2]), double.Parse(extents[3]));
             
                int incrementCount = 0;
                long totalCount = GetTotalTileCount(minAndMaxScales, 256, 256, GeographyUnit.Meter, restrictedExtent);
                //Parallel.ForEach(minAndMaxScalesWithZoom, (scalezoom) =>
                //{
                //    GenerateCacheFiles(layersToCache, scalezoom, restrictedExtent, GeographyUnit.Meter, cacheDirectory, cacheId, ref incrementCount, totalCount);
                //});
                foreach (var scalezoom in minAndMaxScalesWithZoom)
                {
                    GenerateCacheFiles(layersToCache, scalezoom, restrictedExtent, GeographyUnit.Meter, cacheDirectory, cacheId, ref incrementCount, totalCount);
                }
            });
            ButtonText = "Download Current Extent Cache";
            IsEnabled = true;
        }

        private static void Layer_SendingWebRequest(object sender, SendingWebRequestEventArgs e)
        {
            string uri = e.WebRequest.RequestUri.AbsoluteUri + $"&key={Constants.GoogleMapApiKey}"; // set your key in constants file
            e.WebRequest = HttpWebRequest.Create(uri);
        }

        public void GenerateCacheFiles(IEnumerable<Layer> layers, (double, int) scale, RectangleShape restrictedExtent, GeographyUnit mapUnit, string cacheDirectory, string cacheId, ref int counter, long totalCount)
        {
            try
            {
                // Create a MapSuite default tile matrix object.
                TileMatrix matrix = new TileMatrix(scale.Item1, 256, 256, new RectangleShape(-1000000000, 1000000000, 1000000000, -1000000000), mapUnit);

                // Get cells in the specified extent.
                Collection<MatrixCell> cells = matrix.GetIntersectingCells(restrictedExtent);

                // Create a tile cache object with the cache directory and cache id.
                FileRasterTileCache tileCache = new FileRasterTileCache(cacheDirectory, cacheId, RasterTileFormat.Png);

                foreach (MatrixCell cell in cells)
                {
                    // Create a bitmap whose size equals the tile size.
                    using (SKBitmap bitmap = new SKBitmap(256, 256))
                    {
                        // Create an SKCanvas for drawing the passed layers.
                        byte[] imageBytes;
                        var storedBitmap = new GeoImage();
                        using (SKData skData = SKImage.FromBitmap(bitmap).Encode(SKEncodedImageFormat.Png, 100))
                        {
                            imageBytes = skData.ToArray();
                        }

                        using (SKCanvas canvas = new SKCanvas(bitmap))
                        {
                            canvas.Clear(SKColors.Transparent);


                            GeoImage geoImageBitmap = new GeoImage(imageBytes);
                            // Create a GeoCanvas for drawing using SkiaGeoCanvas.

                            SkiaGeoCanvas geoCanvas = new SkiaGeoCanvas();
                            geoCanvas.BeginDrawing(geoImageBitmap, cell.BoundingBox, mapUnit);

                            foreach (Layer layer in layers)
                            {
                                lock (layer)
                                {
                                    if (!layer.IsOpen) { layer.Open(); }
                                    layer.Draw(geoCanvas, new Collection<SimpleCandidate>());
                                }
                            }
                            storedBitmap = geoImageBitmap;
                            geoCanvas.EndDrawing();
                        }

                        // Create GeoImage from SKBitmap
                        GeoImage geoImage = new GeoImage(imageBytes);

                        // Create RasterTile using GeoImage
                        RasterTile tile = new RasterTile(storedBitmap, scale.Item2, cell.Column, cell.Row);

                        // Save tile.
                        tileCache.SaveTile(tile);
                    }
                }
            }
            catch (Exception e)
            {
                var message = e;
            }
        }


        private long GetTotalTileCount(List<double> scales, int tileWidth, int tileHeight, GeographyUnit mapUnit, RectangleShape restrictedExtent)
        {
            List<long> cells = new List<long>();
            foreach (double scale in scales)
            {
                TileMatrix tileMatrix = new TileMatrix(scale, tileWidth, tileHeight, new RectangleShape(-1000000000, 1000000000, 1000000000, -1000000000), GeographyUnit.DecimalDegree);
                RowColumnRange rowColumnRange = tileMatrix.GetIntersectingRowColumnRange(restrictedExtent);
                long deltaColumn = rowColumnRange.MaxColumnIndex - rowColumnRange.MinColumnIndex + 1;
                long deltaRow = rowColumnRange.MaxRowIndex - rowColumnRange.MinRowIndex + 1;
                long cellsCountUnderScale = deltaColumn * deltaRow;
                cells.Add(cellsCountUnderScale);
            }
            long totalTiles = cells.Sum();
            return totalTiles;
        }


        public event PropertyChangedEventHandler PropertyChanged;

        protected void OnPropertyChanged(string propertyName)
        {
            var handler = PropertyChanged;
            if (handler != null)
                handler(this, new PropertyChangedEventArgs(propertyName));
        }
    }
}
